<?php
// lib/roborockvacuum.php

require_once __DIR__ . '/../modules/roborockvacuum/roborockvacuum.class.php';
require_once __DIR__ . '/../modules/roborockvacuum/MiioClient.php';

// === Константы ===
// Мощность всасывания
const ROBOROCK_FAN_POWER_OFF       = 105;
const ROBOROCK_FAN_POWER_SILENT    = 101;
const ROBOROCK_FAN_POWER_STANDARD  = 102;
const ROBOROCK_FAN_POWER_MEDIUM    = 103;
const ROBOROCK_FAN_POWER_TURBO     = 104;
const ROBOROCK_FAN_POWER_MAX       = 108;

// Интенсивность подачи воды
const ROBOROCK_WATER_FLOW_OFF      = 200;
const ROBOROCK_WATER_FLOW_LOW      = 201;
const ROBOROCK_WATER_FLOW_MEDIUM   = 202;
const ROBOROCK_WATER_FLOW_HIGH     = 203;

// Режимы мойки пола
const ROBOROCK_MOP_MODE_STANDARD   = 300;
const ROBOROCK_MOP_MODE_DEEP       = 301;
const ROBOROCK_MOP_MODE_CUSTOM     = 302;
const ROBOROCK_MOP_MODE_DEEP_PLUS  = 303;
const ROBOROCK_MOP_MODE_QUICK      = 304;

// Прочие режимы
const ROBOROCK_DUST_COLLECTION_ON  = 1;
const ROBOROCK_DUST_COLLECTION_OFF = 0;
const ROBOROCK_CARPET_CLEAN_ON     = 1;
const ROBOROCK_CARPET_CLEAN_OFF    = 0;
const ROBOROCK_CHILD_LOCK_ON       = 1;
const ROBOROCK_CHILD_LOCK_OFF      = 0;

// === Внутренние функции ===

/**
 * Инициализация модуля и проверка конфига.
 * @throws Exception
 */
function roborock_init(): \roborockvacuum
{
    static $module;
    if (!$module) {
        $module = new \roborockvacuum();
        $module->getConfig();
    }
    if (empty($module->config['VACUUM_IP']) || empty($module->config['API_TOKEN'])) {
        throw new Exception('Параметры VACUUM_IP или API_TOKEN не заданы.');
    }
    return $module;
}

/**
 * Возвращает инициализированный клиент MiioClient
 * @throws Exception
 */
function roborock_client(): MiioClient
{
    $cfg = roborock_init()->config;
    return new MiioClient(
        $cfg['VACUUM_IP'],
        $cfg['API_TOKEN'],
        false // debug off by default
    );
}

/**
 * Универсальная отправка команды через MiioClient
 * @throws Exception
 */
function roborock_sendRaw(string $method, array $params = [])
{
    $client = roborock_client();
    return $client->send($method, $params);
}

// === Публичные функции ===

/** Получить статус пылесоса */
function roborock_getStatus()
{
    $res = roborock_sendRaw('get_status');
    return is_array($res) && count($res) > 0 ? $res[0] : false;
}

/** Получить уровень звука */
function roborock_getSoundVolume()
{
    $res = roborock_sendRaw('get_sound_volume');
    return is_array($res) && count($res) > 0 ? $res[0] : false;
}

function roborock_changeSoundVolume(int $percent)
{
    return roborock_sendRaw('change_sound_volume', [$percent]);
}

function roborock_testSoundVolume()
{
    return roborock_sendRaw('test_sound_volume');
}

/** Установка режима Не беспокоить */
function roborock_setDndTimer(int $startHour, int $startMin, int $endHour, int $endMin)
{
    return roborock_sendRaw('set_dnd_timer', [$startHour, $startMin, $endHour, $endMin]);
}

function roborock_closeDndTimer()
{
    return roborock_sendRaw('close_dnd_timer');
}

/** Режимы мотора и воды */
function roborock_getCleanMotorMode()
{
    return roborock_sendRaw('get_clean_motor_mode');
}

function roborock_setCleanMotorMode(int $fanPower, int $mopMode, int $waterFlow)
{
    $payload = [
        'fan_power'       => $fanPower,
        'mop_mode'        => $mopMode,
        'water_box_mode'  => $waterFlow,
    ];
    return roborock_sendRaw('set_clean_motor_mode', [$payload]);
}

/** Режимы мойки пола */
function roborock_getMopMode()
{
    return roborock_sendRaw('get_mop_mode');
}

function roborock_setMopMode(int $mode)
{
    return roborock_sendRaw('set_mop_mode', [$mode]);
}

/** Кастомная подача воды */
function roborock_getWaterBoxCustomMode()
{
    return roborock_sendRaw('get_water_box_custom_mode');
}

function roborock_setWaterBoxCustomMode(int $mode)
{
    return roborock_sendRaw('set_water_box_custom_mode', ['water_box_mode' => $mode]);
}

/** Детская блокировка */
function roborock_getChildLockStatus()
{
    return roborock_sendRaw('get_child_lock_status');
}

function roborock_setChildLockStatus(bool $enabled)
{
    $val = $enabled ? ROBOROCK_CHILD_LOCK_ON : ROBOROCK_CHILD_LOCK_OFF;
    return roborock_sendRaw('set_child_lock_status', ['lock_status' => $val]);
}

/** Чистка ковра */
function roborock_getCarpetCleanMode()
{
    return roborock_sendRaw('get_carpet_clean_mode');
}

function roborock_setCarpetCleanMode(bool $enabled)
{
    $val = $enabled ? ROBOROCK_CARPET_CLEAN_ON : ROBOROCK_CARPET_CLEAN_OFF;
    return roborock_sendRaw('set_carpet_clean_mode', ['carpet_clean_mode' => $val]);
}

/** Автосбор пыли */
function roborock_getDustCollectionMode()
{
    return roborock_sendRaw('get_dust_collection_mode');
}

function roborock_setDustCollectionMode(bool $enabled)
{
    $val = $enabled ? ROBOROCK_DUST_COLLECTION_ON : ROBOROCK_DUST_COLLECTION_OFF;
    return roborock_sendRaw('set_dust_collection_mode', ['mode' => $val]);
}

// --- Основные команды уборки ---
function roborock_appStart()
{
    return roborock_sendRaw('app_start');
}

function roborock_appPause()
{
    return roborock_sendRaw('app_pause');
}

function roborock_appStop()
{
    return roborock_sendRaw('app_stop');
}

function roborock_appCharge()
{
    return roborock_sendRaw('app_charge');
}

function roborock_appCleanRooms(array $roomIds)
{
    return roborock_sendRaw('app_segment_clean', $roomIds);
}

// --- Пыль и мойка ---
function roborock_appStartCollectDust()
{
    return roborock_sendRaw('app_start_collect_dust');
}

function roborock_appStopCollectDust()
{
    return roborock_sendRaw('app_stop_collect_dust');
}

function roborock_appStartWash()
{
    return roborock_sendRaw('app_start_wash');
}

function roborock_appStopWash()
{
    return roborock_sendRaw('app_stop_wash');
}

// --- Найти меня ---
function roborock_findMe()
{
    return roborock_sendRaw('find_me');
}

// --- Карты и комнаты ---
function roborock_getRoomMapping()
{
    return roborock_sendRaw('get_room_mapping');
}

/** Получить информацию о расходных материалах */
function roborock_getConsumable()
{
    $res = roborock_sendRaw('get_consumable');
    return is_array($res) && count($res) > 0 ? $res[0] : false;
}

// === Человекочитаемые подсказки ===
/**
 * Возвращает строку статуса по коду, либо false
 */
function roborock_getStatusString(int $code = null)
{
    if ($code === null) {
        $status = roborock_getStatus();
        if (!$status || !isset($status['state'])) {
            return false;
        }
        $code = (int)$status['state'];
    }

    $stateMap = [
        1   => 'Запуск',
        2   => 'Отключено от зарядки',
        3   => 'Ожидание',
        4   => 'Дистанционное управление',
        5   => 'Уборка',
        6   => 'Возвращение на базу',
        7   => 'Ручной режим',
        8   => 'Зарядка',
        9   => 'Проблема с зарядкой',
        10  => 'Пауза',
        11  => 'Точечная уборка',
        12  => 'Ошибка',
        13  => 'Выключение',
        14  => 'Обновление',
        15  => 'Пристыковка',
        16  => 'Переход к цели',
        17  => 'Зональная уборка',
        18  => 'Сегментная уборка',
        22  => 'Опустошение контейнера',
        23  => 'Мойка швабры',
        26  => 'Переход к мойке швабры',
        100 => 'Зарядка завершена',
        101 => 'Устройство офлайн',
    ];

    return $stateMap[$code] ?? false;
}
