<?php
/**
 * Клиент для работы с протоколом miIO
 * Реализует шифрованное UDP-взаимодействие с устройствами Xiaomi
 */
class MiioClient
{
    private string $ip;          // IP-адрес устройства
    private string $token;       // Токен в бинарном формате
    private int $port       = 54321;  // Стандартный порт miIO
    private int $devId      = 0;      // ID устройства
    private int $msgId      = 0;      // Счетчик сообщений
    private int $deviceTs   = 0;      // Временная метка устройства
    private bool $debug;            // Режим отладки

    /**
     * Создание клиента miIO
     * @param string $ip IP-адрес устройства
     * @param string $tokenHex Токен в hex-формате (32 символа)
     * @param bool $debug Включить режим отладки
     * @throws \InvalidArgumentException
     */
    public function __construct(string $ip, string $tokenHex, bool $debug = false)
    {
        if (strlen($tokenHex) !== 32) {
            throw new \InvalidArgumentException("Token должен быть 16 байт в hex");
        }
        $this->ip    = $ip;
        $this->token = hex2bin($tokenHex);
        $this->debug = $debug;
    }

    /**
     * Отправка команды устройству
     * @param string $method Имя метода
     * @param array $params Параметры метода
     * @return array|null Результат выполнения команды
     * @throws \Exception
     */
    public function send(string $method, array $params = []): ?array
    {
        if ($this->devId === 0) {
            $this->discoverDevice();
        }

        $this->msgId++;
        $payload   = json_encode(['id' => $this->msgId, 'method' => $method, 'params' => $params], JSON_UNESCAPED_UNICODE) . "\x00";
        $encrypted = $this->encrypt($payload);

        // Используем счетчик устройства
        $sendTs = $this->deviceTs + 1;

        $length      = 32 + strlen($encrypted);
        $headerNoSig = pack('n', 0x2131)
            . pack('n', $length)
            . pack('N', 0)
            . pack('N', $this->devId)
            . pack('N', $sendTs);

        $sig    = md5($headerNoSig . $this->token . $encrypted, true);
        $packet = $headerNoSig . $sig . $encrypted;

        if ($this->debug) {
            echo "--- MiioClient Debug (send) ---\n";
            echo "Using ts={$sendTs}\n";
            echo "HeaderNoSig: " . bin2hex($headerNoSig) . "\n";
            echo "Checksum:    " . bin2hex($sig) . "\n";
            echo "Encrypted:   " . bin2hex($encrypted) . "\n";
        }

        list($respHeader, $respEncrypted) = $this->udpExchange($packet);

        // Обновляем счетчик устройства из ответа
        $respTsParts    = unpack('Nts', substr($respHeader, 12, 4));
        $this->deviceTs = $respTsParts['ts'];
        if ($this->debug) {
            echo "Got response ts={$this->deviceTs}\n";
        }

        $decrypted = $this->decrypt($respEncrypted);
        if ($this->debug) {
            echo "Decrypted JSON: $decrypted\n";
        }

        $data = json_decode($decrypted, true);
        if (empty($data)) {
            throw new \Exception('Неправильный формат ответа: ' . $decrypted);
        }

        return $data['result'] ?? null;
    }

    /**
     * Обнаружение устройства и получение его ID
     * @throws \Exception
     */
    private function discoverDevice(): void
    {
        if (!function_exists('socket_create')) {
            throw new \RuntimeException('ext-sockets не доступно: отсутствует функция socket_create()');
        }
        // Handshake-пакет: magic=0x2131, len=0x0020, затем 32 байта 0xFF
        $hex    = '21310020' . str_repeat('f', 64);
        $header = hex2bin($hex);

        if ($this->debug) {
            echo "--- MiioClient Debug (discover) ---\n";
            echo "HandshakePkt: " . bin2hex($header) . "\n";
            echo "Unicast to {$this->ip}:{$this->port}\n";
        }

        $sock = socket_create(AF_INET, SOCK_DGRAM, SOL_UDP);
        socket_set_option($sock, SOL_SOCKET, SO_RCVTIMEO, ['sec' => 5, 'usec' => 0]);

        // Отправляем 3 пакета для надежности
        for ($i = 1; $i <= 3; $i++) {
            $sent = socket_sendto($sock, $header, strlen($header), 0, $this->ip, $this->port);
            if ($this->debug) {
                echo "→ Handshake #{$i}: sent {$sent} bytes\n";
            }
        }

        $buf  = '';
        $from = $port = null;
        $len  = @socket_recvfrom($sock, $buf, 1024, 0, $from, $port);
        if ($len === false) {
            $err = socket_last_error($sock);
            $msg = socket_strerror($err);
            socket_close($sock);
            throw new \Exception("socket_recvfrom error ({$err}): $msg");
        }
        if ($this->debug) {
            echo "← Got from {$from}:{$port}, {$len} bytes\n";
            echo "Raw response: " . bin2hex($buf) . "\n";
        }
        socket_close($sock);

        if ($len < 32) {
            throw new \Exception("Слишком короткий ответ ({$len} байт)");
        }

        $respHeader      = substr($buf, 0, 32);
        $parts           = unpack('Ndev', substr($respHeader, 8, 4));
        $tsParts         = unpack('Nts',  substr($respHeader, 12, 4));
        $this->devId     = $parts['dev'];
        $this->deviceTs  = $tsParts['ts'];
        if ($this->debug) {
            echo "Got devId: {$this->devId} (hex: " . dechex($this->devId) . ")\n";
            echo "Got deviceTs: {$this->deviceTs}\n";
            echo "Keeping device IP: {$this->ip}\n";
        }
    }

    /**
     * Обмен UDP-пакетами с устройством
     * @param string $pkt Отправляемый пакет
     * @return array [заголовок ответа, зашифрованные данные]
     * @throws \Exception
     */
    private function udpExchange(string $pkt): array
    {
        if (!function_exists('socket_create')) {
            throw new \RuntimeException('ext-sockets не доступно: отсутствует функция socket_create()');
        }
        $sock = socket_create(AF_INET, SOCK_DGRAM, SOL_UDP);
        socket_set_option($sock, SOL_SOCKET, SO_RCVTIMEO, ['sec' => 5, 'usec' => 0]);

        socket_sendto($sock, $pkt, strlen($pkt), 0, $this->ip, $this->port);
        if ($this->debug) {
            echo "→ Sent {$this->ip}:{$this->port}, " . strlen($pkt) . " bytes\n";
        }

        $buf  = '';
        $from = $port = null;
        $len  = @socket_recvfrom($sock, $buf, 2048, 0, $from, $port);
        if ($len === false) {
            $err = socket_last_error($sock);
            $msg = socket_strerror($err);
            socket_close($sock);
            throw new \Exception("socket_recvfrom error ({$err}): $msg");
        }
        if ($this->debug) {
            echo "← Got from {$from}:{$port}, {$len} bytes\n";
            echo "Raw response hex: " . bin2hex($buf) . "\n";
        }
        socket_close($sock);

        if ($len < 32) {
            throw new \Exception("Слишком короткий ответ ({$len} байт)");
        }

        return [substr($buf, 0, 32), substr($buf, 32)];
    }

    /**
     * Генерация ключа и вектора инициализации для шифрования
     * @return array [ключ, вектор инициализации]
     */
    private function generateKeyIv(): array
    {
        $key = md5($this->token, true);
        $iv  = md5($key . $this->token, true);
        return [$key, $iv];
    }

    /**
     * Шифрование данных
     * @param string $plaintext Исходные данные
     * @return string Зашифрованные данные
     */
    private function encrypt(string $plaintext): string
    {
        if (!function_exists('openssl_encrypt')) {
            throw new \RuntimeException('ext-openssl не доступно: отсутствует функция openssl_encrypt()');
        }
        list($key, $iv) = $this->generateKeyIv();
        return openssl_encrypt($plaintext, 'AES-128-CBC', $key, OPENSSL_RAW_DATA, $iv);
    }

    /**
     * Расшифровка данных
     * @param string $ciphertext Зашифрованные данные
     * @return string Расшифрованные данные
     */
    private function decrypt(string $ciphertext): string
    {
        if (!function_exists('openssl_decrypt')) {
            throw new \RuntimeException('ext-openssl не доступно: отсутствует функция openssl_encrypt()');
        }
        list($key, $iv) = $this->generateKeyIv();
        $decrypted = openssl_decrypt($ciphertext, 'AES-128-CBC', $key, OPENSSL_RAW_DATA, $iv);
        return rtrim($decrypted, "\x00");
    }
}
