<?php
/**
 * RoborockMapParser.php
 *
 * Класс для разбора gzip-файла карты Roborock и рисования базового изображения.
 * Метод drawBaseMap() возвращает GD-ресурс с заливкой комнат их цветами и отрисовкой стен, объектов и пр.
 * Поддерживает PHP 7.x (возвращает ресурс, а не GdImage).
 */

class RoborockMapParser
{
    // Типы блоков
    public const CHARGER                 = 1;
    public const IMAGE                   = 2;
    public const PATH                    = 3;
    public const GOTO_PATH               = 4;
    public const GOTO_PREDICTED_PATH     = 5;
    public const CURRENTLY_CLEANED_ZONES = 6;
    public const GOTO_TARGET             = 7;
    public const ROBOT_POSITION          = 8;
    public const NO_GO_AREAS             = 9;
    public const VIRTUAL_WALLS           = 10;
    public const OBSTACLES               = 13;
    public const OBSTACLES_WITH_PHOTO    = 15;
    public const NO_MOPPING_AREAS        = 12;
    public const BLOCKS                  = 11;
    public const CARPET_MAP              = 17;
    public const MOP_PATH                = 18;
    public const NO_CARPET_AREAS         = 19;
    public const IGNORED_OBSTACLES       = 14;
    public const IGNORED_OBSTACLES_PHOTO = 16;
    public const DIGEST                  = 1024;

    private string $raw;
    private int $width  = 0;
    private int $height = 0;

    /** @var int[][] $pixels  Двумерный массив пикселов (значения байт из IMAGE-блока). */
    private array $pixels    = [];

    /** @var array<int,array{bbox:int[],color:int[]}> $rooms  Bounding box и цвет для каждой комнаты */
    private array $rooms     = [];
    private array $walls     = [];
    private array $paths     = [];
    private array $zones     = [];
    private array $noGoAreas = [];
    private array $obstacles = [];
    private array $charger   = [];
    private array $robot     = [];
    private array $goto      = [];

    /**
     * Конструктор.
     * @param string $gzipData Содержимое файла map.gz (gzip-архив).
     * @throws \RuntimeException если распаковка gzip не удалась.
     */
    public function __construct(string $gzipData)
    {
        $decoded = @gzdecode($gzipData);
        if ($decoded === false) {
            throw new \RuntimeException("Невозможно распаковать gzip-данные");
        }
        $this->raw = $decoded;
    }

    /**
     * Основной метод парсинга: проходит по всем блокам внутри распакованного файла
     * и заполняет внутренние данные (pixels, rooms, walls, obstacles и т.д.).
     */
    public function parse(): void
    {
        // Первые 2 байта после "RR" (0x00–0x01) — всегда "RR",
        // затем 0x02–0x03: длина заголовка. Сохраняем headerLen.
        $headerLen = $this->getUInt16($this->raw, 0x02);
        $pos = $headerLen;
        $totalLen = strlen($this->raw);

        while ($pos < $totalLen) {
            // Каждый блок начинается с:
            // 0x00–0x01: block type (UInt16LE)
            // 0x02–0x03: длина заголовка этого блока (header length)
            $bhLen   = $this->getUInt16($this->raw, $pos + 2);
            $header  = substr($this->raw, $pos, $bhLen);
            $type    = $this->getUInt16($header, 0);
            // Затем, начиная с offset 0x04, идет UInt32LE — длина данных (data length)
            $dataLen = $this->getUInt32($header, 4);
            $data    = substr($this->raw, $pos + $bhLen, $dataLen);

            switch ($type) {
                case self::CHARGER:
                    // Charger location: 4 байта X, 4 байта Y (Int32LE), конвертируются в пикселы
                    $this->charger = $this->mapPointInt32($data, 0);
                    break;

                case self::IMAGE:
                    // Основной IMAGE-блок: содержит данные карты, пиксельные значения и вычисление комнат
                    $this->parseImageBlock($data, $header, $bhLen);
                    break;

                case self::ROBOT_POSITION:
                    // Robot position: 4 байта X, 4 байта Y (Int32LE)
                    $this->robot = $this->mapPointInt32($data, 0);
                    break;

                case self::GOTO_TARGET:
                    // Goto target: 2 байта X, 2 байта Y (UInt16LE)
                    $this->goto = $this->mapPointUInt16($data, 0);
                    break;

                case self::PATH:
                case self::GOTO_PATH:
                case self::GOTO_PREDICTED_PATH:
                    // Пути: первый UInt32LE в header указывает на количество точек * 4 байт каждая
                    $pts = $this->parsePath($header, $data);
                    // Конвертируем каждую точку из мм в пикселы
                    $this->paths[$type] = array_map(
                        fn($p) => [$this->toImg($p[0]), $this->toImg($p[1])],
                        $pts
                    );
                    break;

                case self::CURRENTLY_CLEANED_ZONES:
                    // Cleaned zones: 2 байта count, затем пары (x1,y1,x2,y2) в UInt16LE
                    $this->zones = $this->scaleZones($this->parseZones($header, $data));
                    break;

                case self::NO_GO_AREAS:
                case self::NO_MOPPING_AREAS:
                    // No go / no mop zones: аналогично zones
                    $this->noGoAreas = $this->scaleZones($this->parseZones($header, $data));
                    break;

                case self::VIRTUAL_WALLS:
                    // Virtual walls: пары координат (x1,y1,x2,y2) UInt16LE
                    $this->walls = $this->scaleWalls($this->parseWalls($header, $data));
                    break;

                case self::OBSTACLES:
                case self::OBSTACLES_WITH_PHOTO:
                    // Obstacles: count (UInt16LE), затем каждый obstacle — два Int16LE (x,y)
                    $obs = $this->parseObstacles($header, $data);
                    $this->obstacles = array_map(
                        fn($o) => [$this->toImg($o[0]), $this->toImg($o[1])],
                        $obs
                    );
                    break;

                // Другие типы блоков (BLOCKS, CARPET_MAP и т.д.) пока игнорируем
                default:
                    break;
            }

            // Продвигаем указатель на следующий блок
            $pos += $bhLen + $dataLen;
        }
    }

    /**
     * Парсит IMAGE-блок: заполняет $this->pixels и вычисляет bounding box каждой комнаты.
     *
     * @param string $rawData Бинарные данные пикселей (размер width*height байт)
     * @param string $header  Заголовок блока IMAGE
     * @param int    $hdrLen  Длина заголовка этого блока
     */
    private function parseImageBlock(string $rawData, string $header, int $hdrLen): void
    {
        // В конце заголовка IMAGE-блока (offset hdrLen-8, hdrLen-4) хранятся высота и ширина (UInt32LE)
        $this->height = $this->getUInt32($header, $hdrLen - 8);
        $this->width  = $this->getUInt32($header, $hdrLen - 4);

        // rawData содержит ширину*высоту байт
        // Внутреннее представление: пиксели идут «снизу вверх»,
        // поэтому реальный индекс = x + width * (height - y - 1)
        for ($y = 0; $y < $this->height; $y++) {
            for ($x = 0; $x < $this->width; $x++) {
                $idx = $x + $this->width * ($this->height - $y - 1);
                $pt  = ord($rawData[$idx]);
                $this->pixels[$y][$x] = $pt;

                // Если младшие 3 бита (pt & 0x07) == 7 → это пиксель из комнаты (сканировался)
                if (($pt & 0x07) === 7) {
                    // Старшие биты: ID комнаты = pt >> 3
                    $r = $pt >> 3;
                    if (!isset($this->rooms[$r])) {
                        // Инициализируем bounding box [xmin,ymin,xmax,ymax] и генерируем цвет
                        $this->rooms[$r] = [
                            'bbox'  => [$x, $y, $x, $y],
                            'color' => $this->colorFromRoomId($r),
                        ];
                    } else {
                        // Обновляем bounding box
                        [$xmin, $ymin, $xmax, $ymax] = $this->rooms[$r]['bbox'];
                        $this->rooms[$r]['bbox'] = [
                            min($xmin, $x), min($ymin, $y),
                            max($xmax, $x), max($ymax, $y),
                        ];
                    }
                }
            }
        }
    }

    /**
     * Рисует базовую карту: заливает фон, комнаты, рисует стены, объекты, зоны и проч.
     * Возвращает GD-ресурс.
     *
     * @return resource GD-ресурс, готовый к imagepng()
     */
    public function drawBaseMap()
    {
        // Создаём пустой холст нужного размера
        $img = imagecreatetruecolor($this->width, $this->height);

        // Задаём основные цвета
        $cOut  = imagecolorallocate($img, 240, 240, 240); // Вне дома
        $cWall = imagecolorallocate($img,  50,  50,  50); // Стены
        $cIn   = imagecolorallocate($img, 200, 200, 200); // Внутри
        $cScan = imagecolorallocate($img, 100, 255, 100); // Сканирование
        $cObs  = imagecolorallocate($img, 255,   0,   0); // Препятствия
        $cPath = imagecolorallocate($img,   0,   0, 255); // Путь
        $cGo   = imagecolorallocate($img, 255, 165,   0); // Go-to
        $cZone = imagecolorallocate($img,   0, 255, 255); // Зоны

        // 1) Заливаем фон (вне дома)
        imagefill($img, 0, 0, $cOut);

        // 2) Заливаем bounding box каждой комнаты её цветом
        foreach ($this->rooms as $room) {
            [$xmin, $ymin, $xmax, $ymax] = $room['bbox'];
            [$rr, $gg, $bb] = $room['color'];
            $cRoom = imagecolorallocate($img, $rr, $gg, $bb);
            imagefilledrectangle($img, $xmin, $ymin, $xmax, $ymax, $cRoom);
        }

        // 3) Рисуем пиксельные данные (стены, внутренняя область, скан, препятствия)
        for ($y = 0; $y < $this->height; $y++) {
            for ($x = 0; $x < $this->width; $x++) {
                $pt = $this->pixels[$y][$x];
                switch ($pt) {
                    case 0x00:
                        // Вне дома — оставляем фон
                        break;
                    case 0x01:
                        // Стена
                        imagesetpixel($img, $x, $y, $cWall);
                        break;
                    case 0xFF:
                        // Внутренняя область
                        imagesetpixel($img, $x, $y, $cIn);
                        break;
                    case 0x07:
                        // Сканируемая область
                        imagesetpixel($img, $x, $y, $cScan);
                        break;
                    default:
                        $m = $pt & 0x07;
                        if ($m === 1) {
                            // Стена (альтернативная метка)
                            imagesetpixel($img, $x, $y, $cWall);
                        } elseif ($m !== 7) {
                            // Препятствие
                            imagesetpixel($img, $x, $y, $cObs);
                        }
                        break;
                }
            }
        }

        // 4) Рисуем зоны (filled) и no-go области (контур)
        foreach ($this->zones as $z) {
            imagefilledrectangle($img, $z[0], $z[1], $z[2], $z[3], $cZone);
        }
        foreach ($this->noGoAreas as $z) {
            imagerectangle($img, $z[0], $z[1], $z[2], $z[3], $cObs);
        }

        // 5) Рисуем пути: если PATH, синим; иначе оранжевым (GOTO_PATH и т. д.)
        foreach ($this->paths as $type => $pts) {
            $col = ($type === self::PATH) ? $cPath : $cGo;
            for ($i = 1; $i < count($pts); $i++) {
                imageline(
                    $img,
                    $pts[$i - 1][0], $pts[$i - 1][1],
                    $pts[$i][0],     $pts[$i][1],
                    $col
                );
            }
        }

        // 6) Рисуем препятствия кружками
        foreach ($this->obstacles as $o) {
            imagefilledellipse($img, $o[0], $o[1], 4, 4, $cObs);
        }

        // 7) Зарядка (маленький прямоугольник)
        if ($this->charger) {
            [$cx, $cy] = $this->charger;
            imagefilledrectangle($img, $cx - 3, $cy - 3, $cx + 3, $cy + 3, $cGo);
        }

        // 8) Позиция робота (кружок)
        if ($this->robot) {
            [$rx, $ry] = $this->robot;
            imagefilledellipse($img, $rx, $ry, 6, 6, $cPath);
        }

        // 9) Точка Go-to (крестик)
        if ($this->goto) {
            [$gx, $gy] = $this->goto;
            imageline($img, $gx - 5, $gy, $gx + 5, $gy, $cGo);
            imageline($img, $gx, $gy - 5, $gx, $gy + 5, $cGo);
        }

        return $img;
    }

    /**
     * @return int Ширина карты (пиксели)
     */
    public function getWidth(): int
    {
        return $this->width;
    }

    /**
     * @return int Высота карты (пиксели)
     */
    public function getHeight(): int
    {
        return $this->height;
    }

    /**
     * @return array<int, array{bbox:int[], color:int[]}> Массив комнат:
     *     [ roomId => ['bbox'=>[xmin,ymin,xmax,ymax], 'color'=>[r,g,b]] ]
     */
    public function getRooms(): array
    {
        return $this->rooms;
    }

    // ── Вспомогательные методы для парсинга бинарных данных ──

    /**
     * Преобразует координаты зон из мм → пикселы.
     * @param array<int,int[]> $zones Массив зон [ [x1,y1,x2,y2], ... ] в мм
     * @return array<int,int[]> Массив зон [ [x1',y1',x2',y2'], ... ] в пикселах
     */
    private function scaleZones(array $zones): array
    {
        return array_map(fn($z) => [
            $this->toImg($z[0]), $this->toImg($z[1]),
            $this->toImg($z[2]), $this->toImg($z[3]),
        ], $zones);
    }

    /**
     * Преобразует данные виртуальных стен из мм → пикселы.
     * @param array<int,int[]> $walls Массив стен [ [x1,y1,x2,y2], ... ] в мм
     * @return array<int,int[]> Массив стен в пикселах
     */
    private function scaleWalls(array $walls): array
    {
        return array_map(fn($w) => [
            $this->toImg($w[0]), $this->toImg($w[1]),
            $this->toImg($w[2]), $this->toImg($w[3]),
        ], $walls);
    }

    /**
     * Парсит PATH-блок (блоки пути).
     * @param string $h Заголовок блока (header)
     * @param string $d Данные блока (raw data)
     * @return array<int,int[]> Массив пар координат [ [x,y], [x,y], ... ] в мм
     */
    private function parsePath(string $h, string $d): array
    {
        // getUInt32($h,4) / 4 = количество точек
        $cnt = $this->getUInt32($h, 4) / 4;
        $p   = [];
        for ($i = 0; $i < $cnt; $i++) {
            $o = $i * 4;
            $p[] = [$this->getInt16($d, $o), $this->getInt16($d, $o + 2)];
        }
        return $p;
    }

    /**
     * Парсит зоны (CURRENTLY_CLEANED_ZONES, NO_GO_AREAS и т.п.).
     * @param string $h Заголовок блока
     * @param string $d Данные блока
     * @return array<int,int[]> Массив зон [ [x1,y1,x2,y2], ... ] в мм
     */
    private function parseZones(string $h, string $d): array
    {
        // В заголовке, начиная с offset 8, лежит UInt16LE — число зон
        $cnt = $this->getUInt16($h, 8);
        $z   = [];
        for ($i = 0; $i < $cnt; $i++) {
            $o = $i * 8;
            $z[] = [
                $this->getInt16($d, $o),   $this->getInt16($d, $o + 2),
                $this->getInt16($d, $o + 4), $this->getInt16($d, $o + 6),
            ];
        }
        return $z;
    }

    /**
     * Парсит виртуальные стены (WALLS).
     * @param string $h Заголовок блока
     * @param string $d Данные блока
     * @return array<int,int[]> Массив стен [ [x1,y1,x2,y2], ... ] в мм
     */
    private function parseWalls(string $h, string $d): array
    {
        // В заголовке, начиная с offset 8, лежит UInt16LE — число стен
        $cnt = $this->getUInt16($h, 8);
        $w   = [];
        for ($i = 0; $i < $cnt; $i++) {
            $o = $i * 8;
            $w[] = [
                $this->getInt16($d, $o),   $this->getInt16($d, $o + 2),
                $this->getInt16($d, $o + 4), $this->getInt16($d, $o + 6),
            ];
        }
        return $w;
    }

    /**
     * Парсит OBSTACLES-блок.
     * @param string $h Заголовок блока
     * @param string $d Данные блока
     * @return array<int,int[]> Массив препятствий [ [x,y], ... ] в мм
     */
    private function parseObstacles(string $h, string $d): array
    {
        // В заголовке, начиная с offset 8, лежит UInt16LE — число препятствий
        $cnt = $this->getUInt16($h, 8);
        if ($cnt === 0) {
            return [];
        }
        // Размер одного элемента = длина данных / cnt
        $sz = intdiv(strlen($d), $cnt);
        $o  = [];
        for ($i = 0; $i < $cnt; $i++) {
            $off = $i * $sz;
            $o[] = [$this->getInt16($d, $off), $this->getInt16($d, $off + 2)];
        }
        return $o;
    }

    // ── Методы для конвертации координат и чтения чисел ──

    /**
     * Делит значение в мм на 50 и округляет, возвращая пиксели.
     * @param int $v Значение в мм
     * @return int Значение в пикселах
     */
    private function toImg(int $v): int
    {
        return (int) round($v / 50);
    }

    /**
     * Читает 2-байтовое беззнаковое целое (UInt16LE) из строки $d по смещению $o
     * @param string $d
     * @param int    $o
     * @return int
     */
    private function getUInt16(string $d, int $o): int
    {
        return (ord($d[$o + 1]) << 8) | ord($d[$o]);
    }

    /**
     * Читает 2-байтовое знаковое целое (Int16LE) из строки $d по смещению $o
     * @param string $d
     * @param int    $o
     * @return int
     */
    private function getInt16(string $d, int $o): int
    {
        $v = $this->getUInt16($d, $o);
        return $v & 0x8000 ? $v - 0x10000 : $v;
    }

    /**
     * Читает 4-байтовое беззнаковое целое (UInt32LE) из строки $d по смещению $o
     * @param string $d
     * @param int    $o
     * @return int
     */
    private function getUInt32(string $d, int $o): int
    {
        return
            (ord($d[$o + 3]) << 24) |
            (ord($d[$o + 2]) << 16) |
            (ord($d[$o + 1]) << 8)  |
            ord($d[$o]);
    }

    /**
     * Читает 4-байтовое знаковое целое (Int32LE) из строки $d по смещению $o
     * @param string $d
     * @param int    $o
     * @return int
     */
    private function getInt32(string $d, int $o): int
    {
        $v = $this->getUInt32($d, $o);
        return $v & 0x80000000 ? $v - 0x100000000 : $v;
    }

    /**
     * Читает одну точку из данных (4 байта Int32LE на координату), возвращает [x_px, y_px]
     * @param string $d Данные
     * @param int    $o Смещение
     * @return int[] [x_px, y_px]
     */
    private function mapPointInt32(string $d, int $o): array
    {
        return [
            $this->toImg($this->getInt32($d, $o)),
            $this->toImg($this->getInt32($d, $o + 4))
        ];
    }

    /**
     * Читает одну точку из данных (2 байта UInt16LE на координату), возвращает [x_px, y_px]
     * @param string $d Данные
     * @param int    $o Смещение
     * @return int[] [x_px, y_px]
     */
    private function mapPointUInt16(string $d, int $o): array
    {
        return [
            $this->toImg($this->getUInt16($d, $o)),
            $this->toImg($this->getUInt16($d, $o + 2))
        ];
    }

    private function colorFromRoomId(int $roomId): array
    {
        // Хэшируем ID в md5, берём первые 3 байта и мапим их в диапазон [80…200]
        $hash = md5((string)$roomId);
        $r = 80 + (hexdec(substr($hash, 0, 2)) % 121);  // 0–255→0–120
        $g = 80 + (hexdec(substr($hash, 2, 2)) % 121);
        $b = 80 + (hexdec(substr($hash, 4, 2)) % 121);
        return [$r, $g, $b];
    }
}
