<?php
/**
 * Загрузчик карт Xiaomi
 *
 * Функциональность:
 * 1. Авторизация в Xiaomi Cloud (с поддержкой 2FA)
 * 2. Получение mapName из пылесоса через miIO (UDP 54321)
 * 3. Загрузка .gz файла из облака
 *
 * Требования:
 * - PHP с расширениями curl, openssl и sockets
 */

include_once("MiioClient.php");

chdir(dirname(__FILE__) . '/../../');
$_GET['op'] = 'iframe';

include_once("./config.php");
include_once("./lib/loader.php");
include_once("./lib/threads.php");

define('WAIT_FOR_MAIN_CYCLE', 0);
set_time_limit(0);

include_once("./load_settings.php");

header('X-Accel-Buffering: no');
echo "<html>";
echo "<head>";
echo '<link rel="stylesheet" href="/3rdparty/bootstrap/css/bootstrap.min.css" type="text/css">';
echo '<script type="text/javascript" src="/3rdparty/bootstrap/js/bootstrap.min.js"></script>';
echo "</head>";
echo '<body style="height: auto;overflow: auto;padding: 10px;font-family: Consolas, Verdana;background: #000080;color: #c0c0c0;border-radius: 5px;">';

// === ЗАГРУЖАЕМ КОНФИГ МОДУЛЯ И ПРОВЕРЯЕМ ПАРАМЕТРЫ ===
$module = new roborockvacuum();
$module->getConfig();

// Проверяем наличие всех необходимых параметров
if (empty($module->config['MI_USERNAME'])) {
    echonow("Error: MI_USERNAME is not set in module configuration", 'red');
    exit;
}
if (empty($module->config['MI_PASSWORD'])) {
    echonow("Error: MI_PASSWORD is not set in module configuration", 'red');
    exit;
}
if (empty($module->config['VACUUM_IP'])) {
    echonow("Error: VACUUM_IP is not set in module configuration", 'red');
    exit;
}
if (empty($module->config['API_TOKEN'])) {
    echonow("Error: API_TOKEN is not set in module configuration", 'red');
    exit;
}
if (empty($module->config['REGION_CODE'])) {
    echonow("Error: REGION_CODE is not set in module configuration", 'red');
    exit;
}

// Инициализируем переменные из конфига
$MI_CLOUD_EMAIL = trim($module->config['MI_USERNAME']);
$MI_CLOUD_PASSWORD = trim($module->config['MI_PASSWORD']);
$VACUUM_IP = trim($module->config['VACUUM_IP']);
$VACUUM_TOKEN = trim($module->config['API_TOKEN']);
$COUNTRY = trim($module->config['REGION_CODE']);

// Определяем константы, чтобы функции ниже могли на них ссылаться
define('MI_CLOUD_EMAIL', $MI_CLOUD_EMAIL);
define('MI_CLOUD_PASSWORD', $MI_CLOUD_PASSWORD);
define('VACUUM_IP', $VACUUM_IP);
define('VACUUM_TOKEN', $VACUUM_TOKEN);
define('COUNTRY', $COUNTRY);

// Файл для хранения куки Xiaomi Cloud
define('COOKIE_JAR', __DIR__ . '/xiaomi_cookies.txt');

/**
 * RC4-шифрование/дешифрование данных
 * @param string $base64Key Ключ шифрования в формате base64
 * @param string $data Данные для шифрования/дешифрования
 * @return string Результат шифрования/дешифрования
 */
function rc4_crypt(string $base64Key, string $data): string
{
    $key = base64_decode($base64Key);
    $S = range(0, 255);
    $j = 0;
    $keyLen = strlen($key);

    // KSA
    for ($i = 0; $i < 256; $i++) {
        $j = ($j + $S[$i] + ord($key[$i % $keyLen])) & 0xFF;
        [$S[$i], $S[$j]] = [$S[$j], $S[$i]];
    }

    // PRGA (пропускаем 1024 байта, как в оригинале)
    $i = $j = 0;
    for ($skip = 0; $skip < 1024; $skip++) {
        $i = ($i + 1) & 0xFF;
        $j = ($j + $S[$i]) & 0xFF;
        [$S[$i], $S[$j]] = [$S[$j], $S[$i]];
    }

    // Определяем, пришла ли строка в base64 (тогда дешифруем), иначе шифруем
    $isBase64 = (base64_decode($data, true) !== false && strlen($data) % 4 === 0);
    $bytes = $isBase64 ? base64_decode($data) : $data;
    $res = '';
    $len = strlen($bytes);

    for ($k = 0; $k < $len; $k++) {
        $i = ($i + 1) & 0xFF;
        $j = ($j + $S[$i]) & 0xFF;
        [$S[$i], $S[$j]] = [$S[$j], $S[$i]];
        $K = $S[($S[$i] + $S[$j]) & 0xFF];
        $res .= chr(ord($bytes[$k]) ^ $K);
    }

    return $isBase64 ? $res : base64_encode($res);
}

/**
 * Генерация nonce для авторизации
 * @return string 12 байт: 8 случайных + 4 байта от floor(ms/60000) в base64
 */
function generate_nonce(): string
{
    $ms = floor(microtime(true) * 1000);
    $rand8 = random_bytes(8);
    $timeBytes = pack('N', floor($ms / 60000));
    return base64_encode($rand8 . $timeBytes);
}

/**
 * Вычисление подписанного nonce
 * @param string $nonce Сгенерированный nonce
 * @param string $ssecurity Ключ безопасности
 * @return string Подписанный nonce в формате base64
 */
function compute_signed_nonce(string $nonce, string $ssecurity): string
{
    $raw = base64_decode($ssecurity) . base64_decode($nonce);
    $hash = hash('sha256', $raw, true);
    return base64_encode($hash);
}

/**
 * Генерация подписи ENC для запросов
 * @param string $url URL запроса
 * @param string $method HTTP метод
 * @param string $signedNonce Подписанный nonce
 * @param array $params Параметры запроса
 * @return string Подпись в формате base64
 */
function generate_enc_signature(string $url, string $method, string $signedNonce, array $params): string
{
    $parsed = parse_url($url);
    $path = $parsed['path'];
    if (strpos($path, '/app/') === 0) {
        $path = substr($path, 4);
    }
    $parts = [strtoupper($method), $path];
    foreach ($params as $k => $v) {
        $parts[] = "{$k}={$v}";
    }
    $parts[] = $signedNonce;
    $string = implode('&', $parts);
    $sha1 = sha1($string, true);
    return base64_encode($sha1);
}

/**
 * Выполнение HTTP GET запроса с поддержкой cookies
 * @param string $url URL запроса
 * @param array $headers HTTP заголовки
 * @return array Массив с кодом ответа и телом
 */
function http_get(string $url, array $headers = []): array
{
    $ch = curl_init();
    curl_setopt($ch, CURLOPT_URL, $url);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    // Хранилище куки
    curl_setopt($ch, CURLOPT_COOKIEJAR, COOKIE_JAR);
    curl_setopt($ch, CURLOPT_COOKIEFILE, COOKIE_JAR);
    if (!empty($headers)) {
        curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    }
    curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
    curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 10);
    curl_setopt($ch, CURLOPT_TIMEOUT, 30);
    $body = curl_exec($ch);
    $code = curl_getinfo($ch, CURLINFO_HTTP_CODE) ?: 0;
    curl_close($ch);
    return ['code' => $code, 'body' => $body];
}

/**
 * Выполнение HTTP POST запроса с поддержкой cookies
 * @param string $url URL запроса
 * @param array $fields Поля формы
 * @param array $headers HTTP заголовки
 * @return array Массив с кодом ответа и телом
 */
function http_post(string $url, array $fields = [], array $headers = []): array
{
    $ch = curl_init();
    curl_setopt($ch, CURLOPT_URL, $url);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($ch, CURLOPT_COOKIEJAR, COOKIE_JAR);
    curl_setopt($ch, CURLOPT_COOKIEFILE, COOKIE_JAR);
    curl_setopt($ch, CURLOPT_POST, true);
    curl_setopt($ch, CURLOPT_POSTFIELDS, http_build_query($fields, '', '&'));
    if (!empty($headers)) {
        curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    }
    curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
    curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 10);
    curl_setopt($ch, CURLOPT_TIMEOUT, 30);
    $body = curl_exec($ch);
    $code = curl_getinfo($ch, CURLINFO_HTTP_CODE) ?: 0;
    curl_close($ch);
    return ['code' => $code, 'body' => $body];
}

/**
 * Шаг 1 авторизации: получение _sign
 * @return string|null Значение _sign или null при ошибке
 */
function login_step_1(): ?string
{
    $url = 'https://account.xiaomi.com/pass/serviceLogin?sid=xiaomiio&_json=true';
    $headers = [
        'User-Agent: ' . generate_user_agent(),
        'Content-Type: application/x-www-form-urlencoded'
    ];
    $r = http_get($url, $headers);
    if ($r['code'] !== 200) {
        return null;
    }
    $body = $r['body'];
    $jsonPart = preg_replace('/^&&&START&&&/', '', $body);
    $data = json_decode($jsonPart, true);
    if (json_last_error() !== JSON_ERROR_NONE || empty($data['_sign'])) {
        return null;
    }
    return $data['_sign'];
}

/**
 * Шаг 2 авторизации: проверка учетных данных
 * @param string $sign Значение _sign из шага 1
 * @return array|null Данные авторизации или null при ошибке
 */
function login_step_2(string $sign)
{
    $url = 'https://account.xiaomi.com/pass/serviceLoginAuth2';
    $md5 = strtoupper(md5(MI_CLOUD_PASSWORD));
    $fields = [
        'sid' => 'xiaomiio',
        'hash' => $md5,
        'user' => MI_CLOUD_EMAIL,
        '_sign' => $sign,
        'callback' => 'https://sts.api.io.mi.com/sts',
        'qs' => '%3Fsid%3Dxiaomiio%26_json%3Dtrue',
        '_json' => 'true'
    ];
    $headers = [
        'User-Agent: ' . generate_user_agent(),
        'Content-Type: application/x-www-form-urlencoded'
    ];
    $r = http_post($url, $fields, $headers);
    if ($r['code'] !== 200) {
        return null;
    }
    // Убираем префикс "&&&START&&&"
    $bodyStripped = preg_replace('/^&&&START&&&/', '', $r['body']);
    $data = json_decode($bodyStripped, true);
    if (json_last_error() !== JSON_ERROR_NONE) {
        return null;
    }
    // Если есть notificationUrl → нужен 2FA
    if (!empty($data['notificationUrl'])) {
        return ['twoFactorAuth' => $data['notificationUrl']];
    }
    if (empty($data['ssecurity']) || empty($data['userId']) || empty($data['location'])) {
        return null;
    }
    return [
        'ssecurity' => $data['ssecurity'],
        'userId' => $data['userId'],
        'cUserId' => $data['cUserId'] ?? null,
        'passToken' => $data['passToken'] ?? null,
        'location' => $data['location']
    ];
}

/**
 * Шаг 3 авторизации: получение токенов
 * @param string $locationUrl URL для получения токенов
 * @return bool Результат авторизации
 */
function login_step_3(string $locationUrl): bool
{
    $headers = [
        'User-Agent: ' . generate_user_agent(),
        'Content-Type: application/x-www-form-urlencoded'
    ];
    $r = http_get($locationUrl, $headers);
    return ($r['code'] === 200);
}

/**
 * Полная процедура авторизации
 * @return array|null Данные авторизации или null при ошибке
 */
function login(): ?array
{
    // Сбрасываем/создаём COOKIE_JAR
    @unlink(COOKIE_JAR);
    touch(COOKIE_JAR);

    // Перед первым GET записываем служебные куки sdkVersion и deviceId
    $deviceId = generate_device_id();
    file_put_contents(COOKIE_JAR, ".mi.com\tTRUE\t/\tFALSE\t0\tsdkVersion\taccountsdk-18.8.15\n", FILE_APPEND);
    file_put_contents(COOKIE_JAR, ".xiaomi.com\tTRUE\t/\tFALSE\t0\tsdkVersion\taccountsdk-18.8.15\n", FILE_APPEND);
    file_put_contents(COOKIE_JAR, ".mi.com\tTRUE\t/\tFALSE\t0\tdeviceId\t{$deviceId}\n", FILE_APPEND);
    file_put_contents(COOKIE_JAR, ".xiaomi.com\tTRUE\t/\tFALSE\t0\tdeviceId\t{$deviceId}\n", FILE_APPEND);

    $sign = login_step_1();
    if ($sign === null) {
        return null;
    }

    $step2 = login_step_2($sign);
    if ($step2 === null) {
        return null;
    }
    if (!empty($step2['twoFactorAuth'])) {
        // Нужно открыть в браузере и пройти 2FA
        return ['twoFactorAuth' => $step2['twoFactorAuth']];
    }

    $ssec = $step2['ssecurity'];
    $userId = $step2['userId'];
    $location = $step2['location'];

    if (!login_step_3($location)) {
        return null;
    }

    // Читаем serviceToken из COOKIE_JAR
    $serviceToken = null;
    $cookies = file(COOKIE_JAR, FILE_IGNORE_NEW_LINES | FILE_SKIP_EMPTY_LINES);
    foreach ($cookies as $line) {
        if (strpos($line, "\tserviceToken\t") !== false) {
            $parts = explode("\t", $line);
            $serviceToken = end($parts);
            break;
        }
    }
    if (!$serviceToken) {
        return null;
    }

    return [
        'ssecurity' => $ssec,
        'userId' => $userId,
        'cUserId' => $step2['cUserId'] ?? null,
        'passToken' => $step2['passToken'] ?? null,
        'serviceToken' => $serviceToken,
    ];
}

/**
 * Генерация User-Agent для запросов
 * @return string User-Agent строка
 */
function generate_user_agent(): string
{
    $agentId = '';
    for ($i = 0; $i < 13; $i++) {
        $agentId .= chr(random_int(65, 69));
    }
    return "Android-7.1.1-1.0.0-ONEPLUS A3010-136-{$agentId} APP/xiaomi.smarthome APPV/62830";
}

/**
 * Генерация ID устройства
 * @return string ID устройства
 */
function generate_device_id(): string
{
    $id = '';
    for ($i = 0; $i < 6; $i++) {
        $id .= chr(random_int(97, 122));
    }
    return $id;
}

/**
 * Получение имени карты от пылесоса
 * @param string $ip IP-адрес пылесоса
 * @param string $token Токен авторизации
 * @return string|null Имя карты или null при ошибке
 */
function vacuum_get_map_name(string $ip, string $token): ?string
{
    $port = 54321;
    $timeout = 2.0;

    $tokenBin = hex2bin($token);
    if ($tokenBin === false || strlen($tokenBin) !== 16) {
        return null;
    }

    $payload = json_encode(['id' => 1, 'method' => 'get_map']);
    $payloadLen = strlen($payload);

    // Собираем кадр: 0x2131 + BE(length) + payload JSON
    $frame = "\x21\x31" . pack('n', $payloadLen) . $payload;

    // XOR-шифруем весь кадр по токену
    $xorReq = '';
    for ($i = 0; $i < strlen($frame); $i++) {
        $xorReq .= chr(ord($frame[$i]) ^ ord($tokenBin[$i % 16]));
    }

    // Открываем UDP-сокет
    $sock = socket_create(AF_INET, SOCK_DGRAM, SOL_UDP);
    if ($sock === false) {
        return null;
    }
    socket_set_option(
        $sock,
        SOL_SOCKET,
        SO_RCVTIMEO,
        ['sec' => (int)$timeout, 'usec' => (int)(($timeout - (int)$timeout) * 1000000)]
    );

    // Отправляем
    $sent = @socket_sendto($sock, $xorReq, strlen($xorReq), 0, $ip, $port);
    if ($sent === false) {
        socket_close($sock);
        return null;
    }

    // Ждём ответ
    $buf = '';
    $from = '';
    $portFrom = 0;
    $len = @socket_recvfrom($sock, $buf, 2048, 0, $from, $portFrom);
    socket_close($sock);

    if ($len === false || $len === 0) {
        // Либо таймаут, либо ошибка
        return null;
    }

    // Дешифруем полученные байты тем же XOR
    $dec = '';
    for ($i = 0; $i < strlen($buf); $i++) {
        $dec .= chr(ord($buf[$i]) ^ ord($tokenBin[$i % 16]));
    }
    // Отбрасываем первые 4 байта (0x2131 + length)
    if (strlen($dec) <= 4) {
        return null;
    }
    $json = substr($dec, 4);
    $obj = json_decode($json, true);
    if (json_last_error() !== JSON_ERROR_NONE || empty($obj['result'])) {
        return null;
    }
    if (is_array($obj['result']) && count($obj['result']) > 0) {
        return (string)$obj['result'][0];
    }
    return null;
}

/**
 * Получение имени карты через miIO протокол
 * @param string $ip IP-адрес пылесоса
 * @param string $token Токен авторизации
 * @return string|null Имя карты или null при ошибке
 */
function vacuum_get_map_name_miio(string $ip, string $token): ?string
{
    try {
        $client = new MiioClient($ip, $token);
    } catch (\InvalidArgumentException $e) {
        // Некорректный токен
        return null;
    }

    for ($i = 0; $i < 5; $i++) {
        echonow("Попытка " . ($i + 1) . "<br>");
        try {
            $result = $client->send('get_map_v1');
        } catch (\Exception $e) {
            // Ошибки сети или шифрования
            return null;
        }

        if (empty($result) || !isset($result[0])) {
            return null;
        }

        $raw = $result[0];
        if (strtolower($raw) === 'retry') {
            // сервер попросил повторить
            usleep(500 * 1000);
            continue;
        }

        // Получили имя карты
        return urldecode($raw);
    }

    // По истечении всех попыток – неудача
    return null;
}

/**
 * Выполнение зашифрованного API запроса
 * @param string $url URL запроса
 * @param array $paramsPlain Параметры запроса
 * @param string $ssecurity Ключ безопасности
 * @param string $serviceToken Токен сервиса
 * @param string $userId ID пользователя
 * @return array|null Результат запроса или null при ошибке
 */
function execute_api_call_encrypted(string $url, array $paramsPlain, string $ssecurity, string $serviceToken, string $userId): ?array
{
    // 1) Сгенерировать nonce и signedNonce
    $nonce = generate_nonce();
    $signedNonce = compute_signed_nonce($nonce, $ssecurity);

    // 2) Копируем paramsPlain и добавляем rc4_hash__
    $params = $paramsPlain;
    $params['rc4_hash__'] = generate_enc_signature($url, 'POST', $signedNonce, $params);

    // 3) Шифруем все значения RC4 под signedNonce
    foreach ($params as $k => $v) {
        $params[$k] = rc4_crypt($signedNonce, $v);
    }

    // 4) Пересчитываем signature для зашифрованных значений
    $signature = generate_enc_signature($url, 'POST', $signedNonce, $params);

    // 5) К финальным параметрам POST добавляем signature, ssecurity, _nonce
    $params['signature'] = $signature;
    $params['ssecurity'] = $ssecurity;
    $params['_nonce'] = $nonce;

    // 6) Собираем заголовки и куки
    $headers = [
        'Accept-Encoding: identity',
        'User-Agent: ' . generate_user_agent(),
        'Content-Type: application/x-www-form-urlencoded',
        'x-xiaomi-protocal-flag-cli: PROTOCAL-HTTP2',
        'MIOT-ENCRYPT-ALGORITHM: ENCRYPT-RC4'
    ];
    $cookieStr = sprintf(
        "userId=%s; yetAnotherServiceToken=%s; serviceToken=%s; locale=en_GB; timezone=GMT+02:00; is_daylight=1; dst_offset=3600000; channel=MI_APP_STORE",
        $userId, $serviceToken, $serviceToken
    );

    // 7) Делаем POST
    $ch = curl_init();
    curl_setopt($ch, CURLOPT_URL, $url);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($ch, CURLOPT_POST, true);
    curl_setopt($ch, CURLOPT_POSTFIELDS, http_build_query($params, '', '&'));
    curl_setopt($ch, CURLOPT_HTTPHEADER, array_merge($headers, ["Cookie: {$cookieStr}"]));
    curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
    curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 10);
    curl_setopt($ch, CURLOPT_TIMEOUT, 30);
    $respBody = curl_exec($ch);
    $code = curl_getinfo($ch, CURLINFO_HTTP_CODE) ?: 0;
    curl_close($ch);

    if ($code !== 200 || !$respBody) {
        return null;
    }

    // 8) Дешифруем ответ RC4 (используем signedNonce)
    $decrypted = rc4_crypt($signedNonce, $respBody);
    $json = json_decode($decrypted, true);
    if (json_last_error() !== JSON_ERROR_NONE) {
        return null;
    }
    return $json;
}

/**
 * Получение URL файла карты
 * @param string $mapName Имя карты
 * @param string $country Код страны
 * @param array $loginData Данные авторизации
 * @return string|null URL файла карты или null при ошибке
 */
function get_map_file_url(string $mapName, string $country, array $loginData): ?string
{
    $apiDomain = ($country === 'cn') ? 'api.io.mi.com' : "{$country}.api.io.mi.com";
    $apiUrl = "https://{$apiDomain}/app/home/getmapfileurl";
    $dataJson = json_encode(['obj_name' => $mapName], JSON_UNESCAPED_SLASHES);
    $paramsPlain = ['data' => $dataJson];

    $resp = execute_api_call_encrypted(
        $apiUrl,
        $paramsPlain,
        $loginData['ssecurity'],
        $loginData['serviceToken'],
        $loginData['userId']
    );
    if (empty($resp['result']['url'])) {
        return null;
    }
    return $resp['result']['url'];
}

/**
 * Загрузка данных карты
 * @param string $mapUrl URL файла карты
 * @return string|null Данные карты или null при ошибке
 */
function get_raw_map_data(string $mapUrl): ?string
{
    $ch = curl_init();
    curl_setopt($ch, CURLOPT_URL, $mapUrl);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
    curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 10);
    curl_setopt($ch, CURLOPT_TIMEOUT, 30);
    $data = curl_exec($ch);
    $code = curl_getinfo($ch, CURLINFO_HTTP_CODE) ?: 0;
    curl_close($ch);

    return (($code === 200 && $data !== false) ? $data : null);
}

// =======================================================================
// Главный поток выполнения
// =======================================================================
try {
    echonow('<div><i style="font-size: 7pt;" class="glyphicon glyphicon-info-sign"></i> Начинаем процесс загрузки карты...</div>');

    // Step 1: Login
    echonow('<div><i style="font-size: 7pt;" class="glyphicon glyphicon-info-sign"></i> Авторизация в Xiaomi Cloud...</div>');
    $loginData = login();
    if (!$loginData) {
        echonow("Ошибка авторизации в Xiaomi Cloud", 'red');
        exit;
    }
    echonow("Авторизация успешна", 'green');

    // Step 2: Get map name
    echonow('<div><i style="font-size: 7pt;" class="glyphicon glyphicon-info-sign"></i> Получение имени карты...</div>');
    $mapName = vacuum_get_map_name(VACUUM_IP, VACUUM_TOKEN);
    if (!$mapName) {
        echonow('<div><i style="font-size: 7pt;" class="glyphicon glyphicon-info-sign"></i> Пробуем получить имя карты через miio...</div>');
        $mapName = vacuum_get_map_name_miio(VACUUM_IP, VACUUM_TOKEN);
        if (!$mapName) {
            echonow("Не удалось получить имя карты", 'red');
            exit;
        }
    }
    echonow("Имя карты получено: " . $mapName, 'green');

    // Step 3: Get map URL
    echonow('<div><i style="font-size: 7pt;" class="glyphicon glyphicon-info-sign"></i> Получение URL карты...</div>');
    $mapUrl = get_map_file_url($mapName, COUNTRY, $loginData);
    if (!$mapUrl) {
        echonow("Не удалось получить URL карты", 'red');
        exit;
    }
    echonow("URL карты получен", 'green');

    // Step 4: Download map
    echonow('<div><i style="font-size: 7pt;" class="glyphicon glyphicon-info-sign"></i> Загрузка карты...</div>');
    $mapData = get_raw_map_data($mapUrl);
    if (!$mapData) {
        echonow("Не удалось загрузить карту", 'red');
        exit;
    }

    // Сохраняем карту
    $mapFile = dirname(__FILE__) . '/map.gz';
    file_put_contents($mapFile, $mapData);

    setGlobal("RoborockVacuum1.rooms", json_encode(roborock_getRoomMapping()));

    echonow("Карта успешно загружена и сохранена в " . $mapFile, 'green');
    echonow(
        '<script language="javascript">
        setTimeout(function(){
            window.top.location.href = "' . ROOTHTML . 'admin.php?md=panel&action=roborockvacuum&ok_msg=' . urlencode('Карта успешно загружена') . '";
        }, 5000);
    </script>'
    );


} catch (Exception $e) {
    echonow("Ошибка: " . $e->getMessage(), 'red');
}

echo "</body>";
echo "</html>";
exit;
